from django.core.exceptions import ValidationError

from .validators import call_validator


def mix(*cls):
    return type("", cls, {})


class MinMaxValidationMixin:
    min_value = None
    max_value = None

    def validate_value(self, value):
        value = super().validate_value(value)

        is_none_valid = self.min_value is None and self.max_value is None
        if not is_none_valid and value is None:
            raise ValidationError("Value cannot be None")

        if self.min_value is not None and value < self.min_value:
            raise ValidationError(f"Value cannot be less than {self.min_value}")

        if self.max_value is not None and value > self.max_value:
            raise ValidationError(f"Value cannot be greater than {self.max_value}")
        return value

    def get_help_format(self):
        yield from super().get_help_format()

        if self.min_value is not None:
            yield f" from {self.min_value}"

        if self.max_value is not None:
            yield f" to {self.max_value}"


class PositiveValidationMixin(MinMaxValidationMixin):
    min_value = 0


class CallToPythonMixin:
    call_func = None

    def prepare_python_call(self, value):
        return {"prepared": value}
    
    def get_call_func(self):
        return self.call_func
    
    def python_call(self, *args, **kwargs):
        return self.call_func(*args, **kwargs)
    
    def to_python(self, value):
        value = super().to_python(value)
        prepared_kwargs = self.prepare_python_call(value)

        def _(*args, **kwargs):
            return self.python_call(*args, **prepared_kwargs, **kwargs)
        return _
    
    def get_admin_preview_value(self, value, name):
        first_validator = next((
                v for v in self.get_validators()
                if isinstance(v, call_validator)
            ), None
        )

        if first_validator is None:
            return "No preview validators. Add at least one call_validator"
        
        return first_validator(self.to_python(value))



from django.conf import settings


def get_setting(name, default=None):
    return getattr(settings, "CONTENT_SETTINGS_" + name, default)


try:
    import pkg_resources

    VERSION = pkg_resources.get_distribution("django-content-settings").version
except Exception:
    VERSION = "dev"


CHECKSUM_KEY_PREFIX = (
    get_setting("CHECKSUM_KEY_PREFIX", "CS_CHECKSUM_") + VERSION + "__"
)

USER_TAGS = get_setting(
    "USER_TAGS",
    {
        "favorites": ("⭐", "⚝"),
        "marked": ("💚", "♡"),
    },
)

CACHE_BACKEND = get_setting("CACHE_BACKEND", "default")

CACHE_TIMEOUT = get_setting("CACHE_TIMEOUT", 60 * 60 * 24)

CACHE_SPLITER = get_setting("CACHE_SPLITER", "::")

VALUES_ONLY_FROM_DB = get_setting("VALUES_ONLY_FROM_DB", False)



from threading import local
import hashlib
from functools import lru_cache

from django.core.cache import caches

from .settings import CHECKSUM_KEY_PREFIX, CACHE_SPLITER, CACHE_TIMEOUT, CACHE_BACKEND, VALUES_ONLY_FROM_DB


DATA = local()


def get_cache():
    return caches[CACHE_BACKEND]


@lru_cache(maxsize=None)
def get_cache_key():
    from .conf import ALL

    return calc_checksum({name: ALL[name].version for name in ALL.keys()})


def calc_checksum(values):
    """
    generate md5 hash for a dict with keys and values as strings
    """
    return hash_value(
        CACHE_SPLITER.join(values[name] for name in sorted(values.keys()))
    )


def hash_value(value):
    """
    generate md5 hash for a string
    """
    return hashlib.md5(value.encode("utf-8")).hexdigest()


def set_local_checksum():
    DATA.ALL_VALUES_CHECKSUM = calc_checksum(DATA.ALL_RAW_VALUES)


def push_checksum(value=None, key=None):
    if key is None:
        key = get_cache_key()

    if value is None:
        value = DATA.ALL_VALUES_CHECKSUM

    get_cache().set(CHECKSUM_KEY_PREFIX + key, value, CACHE_TIMEOUT)


def get_checksum_from_store():
    return get_cache().get(CHECKSUM_KEY_PREFIX + get_cache_key())


def get_checksum_from_local():
    return DATA.ALL_VALUES_CHECKSUM


def set_new_value(name, new_value, version=None):
    from .conf import ALL

    prev_value = DATA.ALL_RAW_VALUES.get(name)

    if version is not None and ALL[name].version == version and prev_value != new_value:
        DATA.ALL_RAW_VALUES[name] = new_value
        DATA.ALL_VALUES[name] = ALL[name].to_python(new_value)

    return prev_value


def get_value(name):
    assert DATA.POPULATED
    return DATA.ALL_VALUES[name]


def is_populated():  # better name?
    return getattr(DATA, "POPULATED", False)


def get_db_objects():
    from .models import ContentSetting

    return {v.name: v for v in ContentSetting.objects.all()}


def reset_all_values(trigger_checksum=None):
    if not is_populated():
        DATA.ALL_VALUES = {}
        DATA.ALL_RAW_VALUES = {}
        DATA.ALL_VALUES_CHECKSUM = None
        DATA.POPULATED = False

    from .conf import ALL

    try:
        db = get_db_objects()
    except Exception:
        DATA.POPULATED = False
        return

    DATA.POPULATED = True
    for name in ALL.keys():
        if name in db:
            set_new_value(name, db[name].value, version=db[name].version)
        else:
            if VALUES_ONLY_FROM_DB:
                raise AssertionError(f"VALUES_ONLY_FROM_DB: {name} is not in DB")
            set_new_value(name, ALL[name].default, ALL[name].version)

    assert len(DATA.ALL_VALUES) == len(ALL), "Not all values are populated"

    set_local_checksum()

    if trigger_checksum != get_checksum_from_local():
        push_checksum()


def check_update():
    if not is_populated():
        return

    store_checksum = get_checksum_from_store()
    if store_checksum is not None and store_checksum != get_checksum_from_local():
        reset_all_values(trigger_checksum=store_checksum)


def recalc_checksums():
    """ """

    from .conf import ALL

    db = get_db_objects()

    db_versions = {}
    db_values = {}

    for name in set(ALL.keys()) | set(db.keys()):
        if name in db:
            db_versions[name] = db[name].version
            db_values[name] = db[name].value
        else:
            db_versions[name] = ALL[name].version
            db_values[name] = ALL[name].default

    db_version_key = calc_checksum(db_versions)
    push_checksum(calc_checksum(db_values), db_version_key)

    if db_version_key == get_cache_key() or not is_populated():
        return

    local_values = {}
    for name in ALL.keys():
        if name in db:
            db_version = db[name].version
            if db_version != ALL[name].version:
                local_values[name] = DATA.ALL_RAW_VALUES[name]
            else:
                local_values[name] = db[name].value
        else:
            local_values[name] = ALL[name].default

    push_checksum(calc_checksum(local_values))



from contextlib import contextmanager, ContextDecorator

CONTEXT_DEFAULTS_STACK = []
CONTEXT_DEFAULTS_PROCESSORS_STACK = []


class process:
    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def process(self, kwargs):
        raise NotImplementedError()


class process_set(process):
    def process(self, kwargs):
        return {
            name: self.process_one(
                kwargs.get(name, set()),
                set([value]) if isinstance(value, str) else set(value),
            )
            for name, value in self.kwargs.items()
        }


class set_add(process_set):
    def process_one(self, start, change):
        return start | change


class set_remove(process_set):
    def process_one(self, start, change):
        return start - change


@contextmanager
def context_defaults(*args, **kwargs):
    CONTEXT_DEFAULTS_STACK.append(kwargs)
    CONTEXT_DEFAULTS_PROCESSORS_STACK.append(args)
    try:
        yield
    finally:
        CONTEXT_DEFAULTS_STACK.pop()
        CONTEXT_DEFAULTS_PROCESSORS_STACK.pop()


def context_defaults_kwargs():
    kwargs = {}

    for processors, new_kwargs in zip(
        CONTEXT_DEFAULTS_PROCESSORS_STACK, CONTEXT_DEFAULTS_STACK
    ):
        for processor in processors:
            for name, value in processor.process(kwargs).items():
                kwargs[name] = value

        for name, value in new_kwargs.items():
            kwargs[name] = value

    return kwargs


class content_settings_context(ContextDecorator):
    def __init__(self, **values) -> None:
        super().__init__()
        self.values_to_update = values
        self.prev_values = {}

    def __enter__(self):
        from content_settings.caching import set_new_value

        for name, new_value in self.values_to_update.items():
            self.prev_values[name] = set_new_value(name, new_value)

    def __exit__(self, *exc):
        from content_settings.caching import set_new_value

        for name, new_value in self.prev_values.items():
            set_new_value(name, new_value)



from importlib import import_module
import operator

from collections import defaultdict


from django.conf import settings
from django.apps import apps

from .types.basic import BaseSetting
from .caching import get_value

LAZY_ATTRIBUTE_PREFIX = "lazy__"
CLS_ATTRIBUTE_PREFIX = "type__"

ALL = {}
FETCH_GROUPS = defaultdict(set)


class LazyObject:
    def __init__(self, factory):
        # Assign using __dict__ to avoid the setattr method.
        self.__dict__["_factory"] = factory

    def new_method_proxy(func):
        """
        Util function to help us route functions
        to the nested object.
        """

        def inner(self, *args):
            return func(self._factory(), *args)

        return inner

    def __call__(self, *args, **kwargs):
        return self._factory()(*args, **kwargs)

    __getattr__ = new_method_proxy(getattr)
    __bytes__ = new_method_proxy(bytes)
    __str__ = new_method_proxy(str)
    __bool__ = new_method_proxy(bool)
    __dir__ = new_method_proxy(dir)
    __hash__ = new_method_proxy(hash)
    __class__ = property(new_method_proxy(operator.attrgetter("__class__")))
    __eq__ = new_method_proxy(operator.eq)
    __lt__ = new_method_proxy(operator.lt)
    __le__ = new_method_proxy(operator.le)
    __gt__ = new_method_proxy(operator.gt)
    __ge__ = new_method_proxy(operator.ge)
    __ne__ = new_method_proxy(operator.ne)
    __mod__ = new_method_proxy(operator.mod)
    __hash__ = new_method_proxy(hash)
    __getitem__ = new_method_proxy(operator.getitem)
    __iter__ = new_method_proxy(iter)
    __len__ = new_method_proxy(len)
    __contains__ = new_method_proxy(operator.contains)


for app_config in apps.app_configs.values():
    app = app_config.name
    try:
        content_settings = import_module(app + ".content_settings")
    except ImportError as e:
        if e.name != app + ".content_settings":
            raise
        continue
    for attr in dir(content_settings):
        val = getattr(content_settings, attr)
        if isinstance(val, BaseSetting):
            if attr in ALL:
                raise ValueError("Overwriting content setting {}".format(attr))

            if val.fetch_groups is not None:
                fetch_groups = val.fetch_groups
                if isinstance(fetch_groups, str):
                    fetch_groups = [fetch_groups]
                for fetch_groups in fetch_groups:
                    FETCH_GROUPS[fetch_groups.upper().replace("-", "_")].add(attr)

            ALL[attr] = val


class _Settings:
    def __getattr__(self, name):
        if name.startswith(CLS_ATTRIBUTE_PREFIX):
            return ALL[name[len(CLS_ATTRIBUTE_PREFIX) :]]

        if name.startswith(LAZY_ATTRIBUTE_PREFIX):
            name = name[len(LAZY_ATTRIBUTE_PREFIX) :]
            return LazyObject(lambda: get_value(name))
        return get_value(name)

    def __contains__(self, name):
        if name.startswith(LAZY_ATTRIBUTE_PREFIX):
            name = name[len(LAZY_ATTRIBUTE_PREFIX) :]
        if name.startswith(CLS_ATTRIBUTE_PREFIX):
            name = name[len(CLS_ATTRIBUTE_PREFIX) :]
        return name in ALL


def set_initial_values_for_db(apply=False):
    from content_settings.models import ContentSetting, HistoryContentSetting

    changes = []
    for k in ALL.keys():
        try:
            ContentSetting.objects.get(name=k)
        except ContentSetting.DoesNotExist:
            if apply:
                ContentSetting.objects.create(
                    name=k,
                    value=ALL[k].default,
                    version=ALL[k].version,
                )
                HistoryContentSetting.update_last_record_for_name(k)

            changes.append((k, "create"))

    for cs in ContentSetting.objects.all():
        if cs.name in ALL:
            if cs.version != ALL[cs.name].version:
                if apply:
                    cs.value = ALL[cs.name].default
                    cs.version = ALL[cs.name].version
                    cs.save()
                    HistoryContentSetting.update_last_record_for_name(cs.name)

                changes.append((cs.name, "update"))

        else:
            if apply:
                cs.delete()
                HistoryContentSetting.update_last_record_for_name(cs.name)

            changes.append((cs.name, "delete"))


content_settings = _Settings()



from django.http import HttpResponseNotFound, HttpResponseForbidden, JsonResponse
from content_settings.conf import content_settings

from .conf import ALL, FETCH_GROUPS


def fetch_one_group(request, group_name):
    ret = {}

    for name in FETCH_GROUPS[group_name]:
        setting = ALL[name]
        if not setting.fetch_permission or not setting.fetch_permission(request.user):
            continue
        value = getattr(content_settings, name)
        ret[name] = setting.json_view_value(request, value)

    return JsonResponse(ret)


def fetch_one_setting(request, name):
    settings_name = name.upper().replace("-", "_")
    if settings_name in FETCH_GROUPS:
        return fetch_one_group(request, settings_name)

    try:
        setting = ALL[settings_name]
    except KeyError:
        return HttpResponseNotFound("wrong name")

    if setting.fetch_permission is None:
        return HttpResponseNotFound("hidden")

    if not setting.fetch_permission(request.user):
        return HttpResponseForbidden("permission denied")

    value = getattr(content_settings, settings_name)

    return JsonResponse({settings_name: setting.json_view_value(request, value)})



from itertools import zip_longest

from django.core.exceptions import ValidationError

from .basic import SimpleText
from .mixins import CallToPythonMixin
from .validators import call_validator

class required:
    pass


class SimpleCallTemplate(CallToPythonMixin, SimpleText):
    template_static_data = None
    template_args_default = None
    preview_html = True
    help_format = "Simple <a href='https://docs.djangoproject.com/en/3.2/topics/templates/' target='_blank'>Django Template</a>. Available objects:"

    def prepare_python_call(self, value):
        raise NotImplementedError()
    
    def python_call(self, *args, **kwargs):
        raise NotImplementedError()
    
    def get_template_args_default(self):
        if not self.template_args_default:
            return {}
        return self.template_args_default
    
    def get_template_static_data(self):
        if not self.template_static_data:
            return {}
        return self.template_static_data
    

    def get_help_format(self):
        yield self.help_format
        yield  "<ul>"

        for name in self.get_template_args_default().keys():
            yield f"<li>{name}</li>"
            
        for name in self.get_template_static_data().keys():
            yield f"<li>{name}</li>"

        yield "</ul>"
    
    def prepate_input_to_dict(self, *args, **kwargs):
        template_args_default = self.get_template_args_default()
        no_value={}
        kwargs = {
            **{
                default[0]: (default[1] if arg_value == no_value else arg_value)
                for arg_value, default in zip_longest(
                    args, template_args_default.items(),
                    fillvalue=no_value
                ) if default != no_value
            },
            **kwargs,
        }

        for name, value in kwargs.items():
            if value == required:
                raise ValidationError(f"Missing required argument {name}")
                
        return kwargs
    
    def get_template_static_data(self):
        from django.conf import settings
        from content_settings.conf import content_settings

        ret = {
            "CONTENT_SETTINGS": content_settings,
            "SETTINGS": settings,
        }

        if self.template_static_data is not None:
            ret = {**ret, **self.template_static_data}
        return ret
    
    def get_admin_preview_value(self, value, name):
        ret = super().get_admin_preview_value(value, name)
        if self.preview_html:
            return ret
        return f"<pre>{ret}</pre>"
    

class DjangoTemplate(SimpleCallTemplate):
    def prepare_python_call(self, value):
        from django.template import Template

        return {"template": Template(value.strip())}
    
    def python_call(self, *args, **kwargs):
        from django.template import Context

        template = kwargs.pop("template")
        return template.render(
            Context(
                {
                    **self.get_template_static_data(),
                    **self.prepate_input_to_dict(*args, **kwargs),
                }
            )
        )
    

class DjangoModelTemplate(DjangoTemplate):
    model_queryset = None
    obj_name = "object"

    def get_template_args_default(self):
        return {
            self.obj_name: required,
            **super().get_template_args_default(),
        }
    
    def get_validators(self):
        validators = super().get_validators()

        if self.model_queryset is not None:
            first = self.model_queryset.first()
            if first is not None:
                validators = (*validators, call_validator(self.model_queryset.first()))

        return validators


class SimpleEval(SimpleCallTemplate):
    help_format = "Python code that returns a value"

    def prepare_python_call(self, value):
        return {"template": compile(value, "<string>", "eval")}
    
    def python_call(self, *args, **kwargs):
        template = kwargs.pop("template")
        globs = {
            **self.get_template_static_data(),
            **self.prepate_input_to_dict(*args, **kwargs),
            "__import__": None,
        }

        return eval(template, globs)




from django.core.exceptions import ValidationError

class call_validator:
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
    
    def __call__(self, func):
        try:
            return func(*self.args, **self.kwargs)
        except Exception as e:
            raise ValidationError(str(e))
